(* $Id: godi_exec.mli 530 2005-04-04 11:58:34Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Run commands and collect their output *)

type process_status = Unix.process_status =
    WEXITED of int
  | WSIGNALED of int
  | WSTOPPED of int

type collector =
    [ `Buffer of Buffer.t
    | `Forward of (string -> unit)
    ]

type emitter =
    [ `File of string
    ]

type interactivity =
    { period : float;
        (* how often the callback function is called (seconds) *)
      callback : unit -> bool;
        (* this function is called; if the result is [true], the command
	 * continues to run; if the result is [false], the command is
	 * killed
	 *)
    }

exception Command_not_found of string
  (* Raised if the command cannot be found *)

exception Command_failure of (string * string array * process_status)
  (* Raised if a command does not exit with code 0.
   * Arguments of the exception: (command, command arguments, code).
   *)

exception Command_discontinued
  (* Raised if the callback function returned false. *)


val collect : collector -> string -> unit
  (* Add the string to the collector *)


val call_cmd :
      ?env:(string * string) list ->    (* Change environment *)
      ?cd:string ->                     (* Change directory *)
      ?stdin:emitter ->                 (* Assign stdin *)
      ?stdout:collector ->              (* Assign stdout *)
      ?stderr:collector ->              (* Assign stderr *)
      ?interactivity:interactivity ->   (* Call a function periodly *)
      string * string array -> (* Command + arguments beginning with argv[1] *)
      unit
  (* Call the command with the arguments, and the modifications to the
   * execution environment. stdin is always assigned to /dev/null.
   * If the command fails, the exception Command_failure will be raised.
   *)


val shell : string -> string * string array
  (* The idea is to call
   *   call_cmd(shell "shell command")
   * to run the command from the shell.
   *)


type signal_handler_id

val register_signal_handler : int -> (int -> unit) -> signal_handler_id
  (* Registers another signal handler for the passed signal number.
   * The handler is added to the list of handlers for this signal.
   * Returns an ID which can be used to remove the handler again.
   *
   * The signal handler must not raise exceptions. As a special
   * exception of this rule, it is allowed to raise Sys.Break to force
   * program termination after all handlers have been run. (This
   * is currently done by re-raising Sys.Break.)
   *)

val forget_signal_handler : signal_handler_id -> unit
  (* Removes the signal handler from the list. When there is no
   * more handler for the signal number, the signal handling is
   * reset to the default behaviour.
   *)

(* The module Godi_exec installs signal handlers for:
 * SIGALRM, SIGPIPE: The handler is empty.
 * SIGINT, SIGQUIT, SIGTERM and SIGHUP: The signal is passed to
 *   subprocesses started by Godi_exec. The program terminates.
 * SIGCONT: The signal is passed to subprocesses.
 * SIGTSTP: The signal is passed to subprocesses, and SIGSTOP is
 *   sent to the program itself.
 * SIGCHLD: The subprocesses are scanned, and zombies are removed.
 *
 * There is an at_exit handler that passes SIGTERM to the subprocesses.
 *)

val environment : unit -> string array
  (* Returns the current environment *)


val set_path : string -> unit
  (* Sets the PATH environment variable globally. *)

val set_variable : string -> string -> unit
  (* Sets another environment variable globally *)
