(* $Id: godi_file.ml 495 2004-09-30 21:38:03Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Unix


let catch_unix_error f arg =
  try f arg
  with
      Unix_error(err,_,errarg) ->
	let errarg' = if errarg <> "" then ": " ^ errarg else errarg in
	raise(Sys_error(Unix.error_message err ^ errarg'))
;;


type file_id =
    { file_device : int;
      file_inode : int;
    }

let file_id file =
  let s = stat file in
  { file_inode = s.st_ino;
    file_device = s.st_dev;
  }
;;


let same_file f1 f2 =
  (* Returns [true] if both files exist, and both files are the same *)
  try
    let s1 = stat f1 in
    let s2 = stat f2 in
    s1.st_dev = s2.st_dev && s1.st_ino = s2.st_ino
  with
      Unix_error(_,_,_) -> false
;;


type file_fingerprint =
    { file_mtime : float;
      file_ctime : float;
      file_size : int;
    }
;;

let file_fingerprint f =
  let s = stat f in
  { file_mtime = s.st_mtime;
    file_ctime = s.st_ctime;
    file_size = s.st_size;
  }
;;


let read_dir =
  (* Sys.readdir does not return ".", and "..", so we do not use it *)
  catch_unix_error
    (fun d ->
       let dh = Unix.opendir d in
       let f = ref [] in
       try
	 while true do
	   f := Unix.readdir dh :: !f;
	 done;
	 assert false
       with
	   End_of_file ->
	     Unix.closedir dh;
	     List.rev !f
	 | error ->
	     Unix.closedir dh;   (* Cannot read entry *)
	     raise error
    )
;;


let file_type ?(is_regular = false) 
              ?(is_dir = false) 
	      ?(is_existing = false)
	      ?(is_symlink = false) 
	      ?(is_not_symlink = false) 
	      fn =
  let lkind =
    try
      Some ((Unix.lstat fn).Unix.st_kind) 
    with
	Unix.Unix_error(_,_,_) -> None in
  let kind =
    try
      Some ((Unix.stat fn).Unix.st_kind) 
    with
	Unix.Unix_error(_,_,_) -> None in
  (* Negative logic: *)
  match lkind, kind with
      Some Unix.S_REG, _ -> not is_dir && not is_symlink
    | Some Unix.S_DIR, _ -> not is_regular && not is_symlink
    | Some Unix.S_LNK, Some Unix.S_REG -> not is_dir && not is_not_symlink
    | Some Unix.S_LNK, Some Unix.S_DIR -> not is_regular && not is_not_symlink
    | Some Unix.S_LNK, Some _ -> 
	not is_regular && not is_dir && not is_not_symlink
    | Some Unix.S_LNK, None ->
	not is_regular && not is_dir && not is_existing && not is_not_symlink
    | Some _, _ ->
	not is_regular && not is_dir && not is_symlink
    | None, _ ->
	not is_regular && not is_dir && not is_existing && not is_symlink
;;


let file_exists ?(symlink=false) fn =
  file_type ~is_existing:true fn || (symlink && file_type ~is_symlink:true fn)
;;


let file_is_dir fn =
  file_type ~is_dir:true fn ;;


let file_mtime =
  catch_unix_error
    (fun fn ->
       (Unix.stat fn).Unix.st_mtime)
;;


let rec remove ?(recurse = false) =
  catch_unix_error
    (fun d ->
       if recurse then (
	 if file_type ~is_dir:true ~is_not_symlink:true d then (
	   let l = read_dir d in
	   List.iter
	     (fun f ->
		let df = Filename.concat d f in
		if f <> "." && f <> ".." then (
		  remove ~recurse df
		)
	     )
	     l;
	   Unix.rmdir d
	 )
	 else
	   if file_exists ~symlink:true d then
	     Sys.remove d
       )
       else
	 if file_type ~is_dir:true ~is_not_symlink:true d then
	   Unix.rmdir d
	 else
	   Sys.remove d
    )
;;


let remove_dir ?(recurse = false) fn =
  if file_type ~is_dir:true ~is_not_symlink:true fn then
    remove ~recurse fn
  else
    raise(Sys_error("Not a directory: " ^ fn))
;;


let rec create_dir ?(recurse = false) =
  catch_unix_error
    (fun d ->
       if recurse then (
	 try
	   Unix.mkdir d 0o777
	 with
	     Unix.Unix_error(Unix.EEXIST,_,_) ->
	       (* Only check that it is already a directory: *)
	       if not (file_type ~is_dir:true d) then
		 raise(Sys_error ("Not a directory: " ^ d))
	   | Unix.Unix_error(Unix.ENOENT,_,_) ->
	       (* Try to ensure that the parent exists: *)
	       let d' = Filename.dirname d in
	       create_dir ~recurse:true d';
	       Unix.mkdir d 0o777
       ) else
	 Unix.mkdir d 0o777
    )
;;


let tmpdir_dir = ref None;;
let tmpdir_count = ref 0 ;;

type tmpdir = string ;;

let tmpdir_path t = t ;;

let rec create_tdir prefix n =
  try
    let name = prefix ^ string_of_int n in
    Unix.mkdir name 0o777;
    name
  with
      Unix.Unix_error(Unix.EEXIST,_,_) ->
	create_tdir prefix (n+1)
;;


let create_tmp_dir() =
  let tmp_prefix = "/tmp/godi" ^ string_of_int (Unix.getpid()) ^ "-" in
  catch_unix_error (create_tdir tmp_prefix) 0
;;


let allocate_tmpdir() =
  (* Ensure tmpdir_dir exists: *)
  let tmpdir_loc =
    match !tmpdir_dir with
	None ->
	  let tmp_prefix = "/tmp/godi" ^ string_of_int (Unix.getpid()) ^ "-" in
	  let d = catch_unix_error (create_tdir tmp_prefix) 0 in
	  tmpdir_dir := Some d;
	  d
      | Some d -> d
  in
  (* Fetch next number: *)
  let n = !tmpdir_count  in
  incr tmpdir_count;
  (* Create directory: *)
  let name = Filename.concat tmpdir_loc (string_of_int n) in
  catch_unix_error (Unix.mkdir name) 0o777;
  name
;;


let release_tmpdir dir =
  remove_dir ~recurse:true dir
;;


let release_all_tmpdirs() =
  match !tmpdir_dir with
      None -> ()
    | Some d ->
	remove_dir ~recurse:true d;
	tmpdir_dir := None
;;


let copy ?(prefer_link = false) ?(remove = false) origname copyname =
  
  let do_copy =
    catch_unix_error
      (fun () ->
	 let b = String.create 4096 in
	 let origfd = open_in_bin origname in
	 try
	   if remove && Sys.file_exists copyname then
	     Sys.remove copyname;
	   let copyfd = open_out_bin copyname in
	   ( try
	       let eof = ref false in
	       while not !eof do
		 let n = input origfd b 0 (String.length b) in
		 if n=0 then eof := true;
		 output copyfd b 0 n
	       done;
	       let p = 
		 (Unix.fstat (Unix.descr_of_in_channel origfd)).Unix.st_perm in
	       Unix.fchmod (Unix.descr_of_out_channel copyfd) (p land 0o777);
	       close_in origfd;
	       close_out copyfd;
	     with
		 error ->
		   close_out copyfd; raise error
	   )
	 with
	     error -> 
	       close_in origfd; raise error
      )
  in

  (* Try to link, if demanded and possible: *)
  if prefer_link then
    try
      Unix.link origname copyname
    with
	Unix.Unix_error((Unix.EXDEV | Unix.EPERM | Unix.EMLINK), _, _) ->
	  do_copy()
      | Unix.Unix_error(_,_,_) as e ->
	  catch_unix_error (fun () -> raise e) ()
      | Invalid_argument _ ->
	  (* [link] not available *)
	  do_copy()
  else
    do_copy()
;;


let execute ?(setx=false) cmd =
  try
    let switches =
      if setx then [ "-xc" ] else [ "-c" ] in
    Unix.execv "/bin/sh" (Array.of_list
			    (["/bin/sh"] @ switches @ [cmd]));
    (* Note: Some old shells do not accept /bin/sh -x -c <cmd>, but only
     * /bin/sh -xc <cmd>!
     *)
  with
      Unix.Unix_error(err, _, param) ->
	let s = Unix.error_message err in
	let s' = if param="" then s else s ^ ": " ^ param in
	raise(Sys_error s')
;;
